###PERL;

# bot is generated from bot.1.pl
# 13.01.2016
#
# This is the facebook bot. It depends on the proxy.
# It reads pages from m.facebook.com archived on the proxy, extracts threads,
# posts, images, etc. from a facebook group and saves them.
#
#    Copyright (C) 2015-2016  Balthasar Szczepański
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

use strict;
use Fcntl;
use File::Copy;
###PROXY_LIB;
###FACEBUG_LIB;
use proxy_lib qw(url2path urldiv2path path2urldiv getcgi divideurl joinurl readconfigfile entitydecode entityencode urldecode urlencode readheaderfile);
use facebug_lib qw(key readdatafile writedatafile gettimenumber);
use POSIX qw(strftime locale_h);

###ARCH_PATH;
###GROUPSETTINGS_PATH;
###KEY_BITS;
###MAX_REDIRECTIONS;
###DETAILED_LOGS;

my $time = time();
srand ($time-$$);
unless (setlocale (LC_TIME, "C")) {
	print "Can't set time locale!\n\n";
	exit;
}

print strftime("%d.%m.%Y %H:%M:%S", gmtime($time))."\n";
# If there are commandline arguments the bot will only process the facebook
# group(s) defined by them. The arguments are numbers which are the group's ID
# and also the name of the config file.
if (scalar @ARGV) {
	for (my $ind=0; $ind < scalar @ARGV; ++$ind) {
		processgroup($ARGV[$ind]);
	}
}
# If there were no commandline arguments the bot will process all groups which
# have a config file defined.
else {
	my $dir;
	my $subpath;
	if (opendir ($dir, GROUPSETTINGS_PATH)) {
		while (defined ($subpath = readdir $dir)) {
			processgroup($subpath);
		}
		closedir ($dir);
	}
}

# The function to process one facebook group. The argument is a number which is
# the group's ID and also the config file name.
sub processgroup {
	(my $filenumber) = @_;
	my $settingspath;
	my %settings;
	my $groupid;
	my $namespath;
	my %names;
	my $archpath;
	
	# ID must be a number
	if ($filenumber =~ /^([0-9]+)$/) {
		$settingspath = GROUPSETTINGS_PATH.$1;
	}
	else {
		return;
	}
	
	# read the configuration file
	%settings = readconfigfile($settingspath);
	# The settings defined so far:
	# id - the group's ID. SHOULD be the same as the file name.
	# hidenames - defines if the real names of people will be hidden or converted
	#             "0" or "no" means no, anything else means yes. default value if
	#             undefined is yes. Not implemented yet (maybe never will). Now it
	#             will always hide names.
	
	if ($settings{'id'} =~ /^([0-9]+)$/) {
		$groupid=$1;
	}
	else {
		return;
	}
	
	#read the list of names. A default name MUST be defined.
	$namespath = GROUPSETTINGS_PATH.$groupid.'-names';
	%names = readconfigfile($namespath);
	if ($names{'default'} eq '') {
		return;
	}
	
	print "Group $groupid\n";
	# process the pages with threads:
	$archpath = url2path('https://m.facebook.com/groups/'.$groupid);
	processdir($archpath.'@q/',\%settings,\%names,0);
	processdir($archpath.'/@q/',\%settings,\%names,0);
	# process the pages with posts:
	$archpath = url2path('https://m.facebook.com/comment/replies');
	processdir($archpath.'@q/',\%settings,\%names,1);
	processdir($archpath.'/@q/',\%settings,\%names,1);
	
}

# The function to process all files in a directory and (recursively) all
# subdirectories.
# The first argument is the directory path (ending with "/").
# The three other arguments are passed to processfile().
sub processdir {
	(my $dirpath, my $settings, my $names, my $pagemode) = @_;
	my $dir;
	my $subpath;
	my $subpathfull;
	print "Dir $dirpath\n";
	
	unless ( opendir ($dir, $dirpath)) {
		return;
	}
	while (defined($subpath = readdir $dir)) {
		$subpathfull=$dirpath.$subpath;
		# "." or ".." should be ignored 
		if ($subpath =~ /^\.\.?$/) {
			next;
		}
		if (-f $subpathfull) {
			processfile($subpathfull, $settings, $names, $pagemode);
		}
		elsif (-d $subpathfull) {
			processdir($subpathfull.'/', $settings, $names, $pagemode);
		}
	}
	closedir ($dir);
}

# The function to process one archived page.
# Arguments:
# 1. The path to the archived header. Ends with "@h". Otherwise function just
#    returns without doing anything.
# 2. reference to the hash with group settings
# 3. reference to the hash with names
# 4. mode, determines how pages are interpreted. If 0 page type can be
#    'thread' - one thread
#    'group' - list of threads' firstposts
#     If nonzero page type can be
#     'post' - one post (with replies)
#
sub processfile {
	(my $headerpath, my $settings, my $names, my $pagemode) = @_;
	my $basepath;
	my $contentpath;
	my $contentfile;
	
	my %header;
	
	my $prot;
	my $host;
	my $port;
	my $path;
	my $query;
	
	my %cgi;
	
	my $postid;     #id of post
	my $threadid;   #id of thread
	my $groupid=0;  #determines if group id was found on page. REAL group id is in $$settings{'id'}!
	my $timenumber; #number, determines when page was saved on the proxy.
	
	my %thread;  #thread info, created by interpreting the page
	my %thread2; #thread info, created by reading previously archived file
	
	my %post;    #post info, created by interpreting the page
	my %post2;   #post info, created by reading previously archived file
	
	my $pagetype; #type of page: 'group', 'thread' or 'post'
	
	
	# Argument must be HEADER path.
	if ($headerpath =~ /^((.+)\@h)$/) {
		$headerpath = $1;
		$basepath = $2;
		$contentpath = $basepath.'@v';
	}
	else {
		return;
	}
	
	($prot, $host, $port, $path, $query) = path2urldiv($basepath);
	
	print 'Page '.joinurl($prot, $host, $port, $path, $query)."\n";
	
	#Determine what type of page it is. If none ot the three types - return.
	if($query ne '') {
		%cgi=getcgi($query);
		if($pagemode) {
			$postid = $cgi{'ctoken'};
			if ($postid =~ /^([0-9]+)_([0-9]+)$/) {
				$threadid = $1;
				$postid = $2;
				$pagetype = 'post';
			}
			else {
				return;
			}
		}
		else{
			$threadid = $cgi{'id'};
			if ($threadid =~ /^([0-9]+)$/) {
				$threadid = $1;
				$pagetype='thread';
			}
			else {
				$pagetype = 'group';
			}
		}
	}
	else {
		if($pagemode) {
			return;
		}
		else {
			$threadid='';
			$pagetype = 'group';
		}
	}
	print "type=$pagetype\n";
	
	# Read the http header. Only interested in the status. If a redirection then
	# follow it and read again. If 200 it's ok to continue processing. Otherwise
	# return.
	for (my $ind=0; $ind<MAX_REDIRECTIONS; ++$ind) {
		%header = readheaderfile($headerpath); 
		if ($header{'status'} =~ /^30[1237] /) {
			my $location;
			unless (defined($location = $header{'location'})) {
				return;
			}
			if ($location !~ /^[a-z]+:\/\//) {
				$location = $prot.'://'.$host.(($port ne '')?(':'.$port):'').$location;
			}
			($prot, $host, $port, $path, $query) = divideurl($location);
			$basepath = urldiv2path($prot, $host, $port, $path, $query);
			$headerpath = $basepath.'@h';
			$contentpath = $basepath.'@v';
		}
		else {
			last;
		}
	}
	if ($header{'status'} !~ /^200 /) {
		return;
	}
	
	# find out when the page was archived on the proxy.
	unless ($timenumber = gettimenumber ($header{'date'})) {
		my @stat;
		unless (@stat = stat $contentpath) {
			$timenumber='00000000000000';
		}
		else {
			$timenumber = strftime('%Y%m%d%H%M%S',gmtime($stat[9]));
		}
	}
	
	# Now the bot can start reading the file
	
	unless (open ($contentfile, "<",$contentpath)) {
		print "Can't open $contentpath.\n";
		return;
	}
	
	my $text;      # piece of text read from the file
	my %tag;       # html tag
	my $mode;      # the state of the main state machine
	my $level;     # keeps track in how many <div> levels the bot went
	my $level2;    # same as above used in different state (previous must be kept)
	
	my $level3;
	my $level4;
	my $shared;
	my $attnumber2;
	
	my $closetag=0;# if there is a tag to close
	my $ignoretext;# if text should be ignored and not added to post/thread content
	my $link;      # if bot is inside a link
	my $namelink;
	my $hidename;  # if bot is inside a part which contains a name to be hidden
	my $attnumber; # number of current attachment
	my $incomplete;# if thread firstpost's content is incomplete 
	my $firstpost; # if the bot is in the firstpost (important if pagetype='post')
	my $description;
	
	my $content;
	
	# Set initial values depending on page type.
	if ($pagetype eq 'thread') {
		print "Thread $threadid\n";
		
		$thread{'id'}=$threadid;
		$thread{'groupid'}=$$settings{'id'};
		$thread{'timenumber'}=$timenumber;
		$mode = 'threads';
		# $level=0;
		# $attnumber=0;
		# $incomplete=0;
		# $shared=0;
	}
	elsif ($pagetype eq 'post'){
		print "Post $postid ($threadid)\n"; 
		
		$mode='posts';
		$firstpost=1;
	}
	else { #group
		print "Threads\n";
		$mode = 'threads';
	}
	
	my $line;
	
	# The main loop. It reads the file like this:
	# <tag>text<tag>text<tag>text...
	# If there two tags next to each other <tag1><tag2> there still is text
	# between them, only with zero length.
	# It the tag looks like this:
	# <tag name1="value1" />
	# the bot will interpret it as
	# <tag name1="value1"></tag>
	# without the 0-length text between.
	#
	# This is a state machine
	# In the main loop the tag is read and depending on the tag's content, state
	# and some variable values some actions can be taken.
	# Then the text is read and again depending ot the text, state and some
	# variable values some actions can be taken.
	# This loop continues until the end of file.
	
	#read and ignore text before the first tag.
	local $/ = '<';
	unless (defined ($text = <$contentfile>)) {
		close($contentfile);
		return;
	}
	# main loop
	while ($mode ne '') {
		# if there was a tag ending with "/>" in the previous iteration then in this
		# one it will be treated as the same tag but with "</".
		if ($closetag){
			$tag{'<'} = '/'.$tag{'<'};
			$tag{'/'}='/';
			delete $tag{"\\"};
			$closetag=0;
		}
		
		# otherwise read next tag
		else {
			local $/ = '>';
			unless (defined ($text = <$contentfile>)) {
				close($contentfile);
				return;
			}
			$text =~ s/>$//;
			# # DEBUG:
			# if($pagetype eq 'thread'){
				# print ">>$mode: <$text>\n";
			# }
			
			# get the attributes of the tag
			# special values: '<' - tag name (with or without '/'), '/' - when it's a
			# closing tag, '\' when tag ends with "/>".
			%tag = taginfo($text);
		}
		local $/ = "\n";
		
		# List of threads. Look for <div>s with threads
		if ($mode eq 'threads'){
			# Thread found. Id not known yet!
			if (($tag{'<'} eq 'div') and ($tag{'id'} =~ /^([a-zA-Z0-9]_[a-zA-Z0-9]_[a-zA-Z0-9])$/)) {
				$mode = 'thread';
				$content=0;
				# set initial values
				if($pagetype ne 'thread') {
					print "Thread [$1]\n";
					%thread = ();
					$thread{'groupid'}=$$settings{'id'};
					$thread{'timenumber2'}=$timenumber;
				}
				$level = 0;
				$attnumber=0;
				$incomplete=0;
				$shared=0;
			}
		}
		
		# one thread
		elsif ($mode eq 'thread'){
			# thread author is in first <h3>
			if (($tag{'<'} eq 'h3') and ($thread{($shared?'shared-':'').'author'} eq '')) {
				$mode = 'thread-author';
				$link=0;
				$namelink=0;
				unless ($shared) {
					$description=0;
				}
				$hidename;
			}
			# elsif (($tag{'<'} eq 'div')and($tag{'id'} !~ /^ufi_/)) {
			elsif ($tag{'<'} eq 'div') {
				# Post content is (always?) in the first <div> with a 2 letter class name after author
				if (($tag{'class'} =~ /^[a-z]{2}$/) and (!defined($thread{($shared?'shared-':'').'postcontent'})) and (defined($thread{($shared?'shared-':'').'author'}))) {
					$mode='thread-content';
					$content=1;
					$level2=0;
					$ignoretext=1; # text in firstposts only inside <p>
					$hidename=0;
					$link=0;
					$namelink=0;
				}
				# The thread contains "shared" content
				elsif (($tag{'id'} =~ /^([a-zA-Z0-9]_[a-zA-Z0-9]_[a-zA-Z0-9])$/) and !$shared and ($thread{'shared'} eq '')) {
					$shared=1;
					$attnumber2=$attnumber;
					$level3=$level;
					$level4=$level2;
					$attnumber=0;
					$level=0;
					$thread{'shared'}=1;
				}
				else {
					++$level;
				}
			}
			elsif (($tag{'<'} eq '/div') and $level) {
				--$level;
			}
			
			# the time text is in the only <abbr> in thread firstpost
			elsif ($tag{'<'} eq 'abbr') {
				$mode = 'thread-time';
			}
			
			elsif ($tag{'<'} eq 'a') {
				if (lc($tag{'aria-label'})eq 'likes') {
					$mode='thread-likes';
				}
				### this code had to be duplicated:
				# there is an image attached
				elsif ($tag{'href'} =~ /^\/photo\.php\?(.*&)?fbid=([0-9]+)(&.*)?$/) {
					my $imgnum=$2;
					my $imgid='a_';
					
					while(length($imgnum)>240) {
						$imgid.=substr($imgnum,0,120).'-/';
						$imgnum=substr($imgnum,120);
					}
					$imgid.=$imgnum;
					
					++$attnumber;
					$thread{($shared?'shared-':'').'img-'.$attnumber}=$imgid;
					$mode = 'thread-attachment-img';
				}
				elsif ($tag{'href'} =~ /^\/[A-Za-z0-9\.]+\/photos\/([^\?]+)(\?.*)?$/) {
					my $imgurl = urldecode($1);
					my $imgid = 's_';
					$imgurl =~ s/([^A-Za-z0-9_\.])/sprintf ("@%02X",ord($1))/eg;
					
					while(length($imgurl)>240) {
						$imgid.=substr($imgurl,0,120).'-/';
						$imgurl=substr($imgurl,120);
					}
					$imgid.=$imgurl;
					
					++$attnumber;
					$thread{($shared?'shared-':'').'img-'.$attnumber}=$imgid;
					$mode = 'thread-attachment-img';
				}
				elsif ($tag{'href'} =~ /^\/video_redirect\/?\?(.*)$/) {
					my %linkcgi = getcgi($1);
					my $imgurl;
					my $imgid='v_';
					if ($linkcgi{'src'} =~ /^https?:\/\/([a-z0-9\.\-]+)?fbcdn\.net\/(hvideo[^\?]+)(\?.*)?$/) {
						$imgurl= $2;
						$imgurl =~ s/([^A-Za-z0-9_\.])/sprintf ("@%02X",ord($1))/eg;
						
						while(length($imgurl)>240) {
							$imgid.=substr($imgurl,0,120).'-/';
							$imgurl=substr($imgurl,120);
						}
						$imgid.=$imgurl;
						++$attnumber;
						$thread{($shared?'shared-':'').'img-'.$attnumber}=$imgid;
						$mode = 'thread-attachment-img';
					}
				}
				###END OF DUPLICATED CODE
				# there is a link attached
				elsif ($tag{'href'} =~ /^https?:\/\/([a-z0-9\.\-]+)?facebook\.com\/l\.php\?(.*&)?u=([^&]+)(&.*)?$/) {
					++$attnumber;
					$thread{($shared?'shared-':'').'link-'.$attnumber}=urldecode($3);
					$mode = 'thread-attachment-link';
				}
				# if thread id is not known it can be determined from this link.
				# also, the number of replies may be found in one of these links.
				elsif ($tag{'href'} =~ /^\/groups\/$$settings{'id'}\/?\?(.*&)?id=([^&]+)(&.*)?$/) {
					if ($thread{'id'} eq '') {
						$thread{'id'} = $2;
						print "Thread $thread{'id'}\n";
					}
					$mode = 'thread-replies';
				}
			}
			
			# Depending on page mode the thread and firstpost information can end in
			# two different ways. When page type is 'thread' the end is at the <div>
			# whose id starts with "ufi_". When page type is 'group' it ends when
			# leaving the thread-related div
			# 
			# elsif ((($tag{'<'} eq 'div') and ($tag{'id'} =~ /^ufi_/))or(($tag{'<'} eq '/div') and ($level ==0))) {
			elsif (($tag{'<'} eq '/div') and ($level ==0)) {
				if($shared){
					$shared=0;
					$attnumber=$attnumber2;
					$level=$level3;
					$level2=$level4;
				}
				else {
					# depending on page type the rest of the page contains posts or other
					# threads.
					if ($pagetype eq 'thread') {
						$mode='posts';
					}
					else {
						$mode='threads';
					}
					
					# Now prepare to save the file 
					my $threadfile;
					my $threadpath = ARCH_PATH.$$settings{'id'}.'/';
					unless (-d $threadpath) {
						unless (mkdir $threadpath) {
							print "Can't mkdir $threadpath.\n";
						}
					}
					$threadpath.='thread/';
					unless (-d $threadpath) {
						unless (mkdir $threadpath) {
							print "Can't mkdir $threadpath.\n";
						}
					}
					$threadpath.=$thread{'id'};
					
					if (sysopen ($threadfile, $threadpath, O_RDWR | O_CREAT)) {
						if (flock ($threadfile, 2)) {
							# read the data already saved in file
							%thread2 = readdatafile($threadfile);
							
							# in 'threads' page type the firstpost's content may be incomplete.
							# in that case it should not be written to file if there is already
							# one. Even if the file has an older version.
							# but in this page type there is information about the number of
							# replies, not found in 'thread' page type. This information should
							# be written even when the post contnet shouldn't.
							#
							# That's why a thread has the timenumber and timenumber2.
							# timenumber defines the time of the post content. in pagetype
							# 'thread' only timenumber is checked and only timenumber is
							# updated.
							# timenumber2 defines the time of information only available in the
							# 'group' page type. In this pagetype the post content is only
							# updated if complete and timenumber allows it. Other information is
							# updated if timenumber2 allows it 
							
							# Don't overwrite newer information with older.
							if ((($pagetype eq 'thread')and($thread2{'timenumber'} ne '')and($thread2{'timenumber'}>=$thread{'timenumber'}))or(($pagetype ne 'thread')and($thread2{'timenumber2'} ne '')and($thread2{'timenumber2'}>=$thread{'timenumber2'}))) {
								print ("Newer version already saved.\n\n");
							}
							else {
								if($pagetype ne 'thread'){
									# Don't overwrite newer post content with older.
									if(($thread2{'timenumber'} ne '')and($thread2{'timenumber'} > $thread{'timenumber2'})) {
										print ("Newer version of post content already saved.\n");
										delete $thread{'postcontent'};
									}
									elsif($incomplete) {
										# Don't overwrite complete post content with incomplete one.
										# Write incomplete content if nothing was archived before,
										# better this than nothing.
										print ("Post content incomplete.\n");
										if(defined($thread2{'postcontent'})){
											delete $thread{'postcontent'};
										}
									}
									else {
										$thread{'timenumber'}=$thread{'timenumber2'};
									}
								}
								# delete previous information about attachments - the numbers
								# could have changed.
								foreach my $ind (keys %thread2) {
									if($ind =~ /^(shared-)?((img(key)?)|(link(text|title)?))-[0-9]+$/) {
										delete $thread2{$ind};
									}
								}
								
								if ($description <= 0) {
									delete $thread{'description'};
								}
								
								# overwrite previous information with new one
								foreach my $ind (keys %thread) {
									$thread2{$ind}=$thread{$ind};
								}
								if ($thread2{'key'} eq '') {
									$thread2{'key'} = key(KEY_BITS);
								}
								
								# write data to file
								if (seek($threadfile, 0, 0)) {
									writedatafile($threadfile,%thread2);
									truncate ($threadfile , tell($threadfile));
									
									if (DETAILED_LOGS) {
										foreach my $ind (keys %thread2) {
											print "$ind: $thread2{$ind}\n"; ####
										}
									}
									print "saved.\n\n";
								}
								else {
									print "Can't seek $threadfile.\n\n";
								}
							}
						}
						else {
							print "Can't lock $threadfile.\n\n";
						}
						close ($threadfile);
					}
					else
					{
						print "Can't open $threadpath.\n\n";
					}
				}
			}
		}
		
		# author name
		elsif ($mode eq 'thread-author') {
			# name can be found in hyperlinks
			if ($tag{'<'} eq 'span') {
				$description=-1;
			}
			elsif ($tag{'<'} eq 'a') {
				if ($tag{'href'} =~ /^https?:\/\/([a-z0-9\.\-]+)?facebook\.com\/l\.php\?(.*&)?u=([^&]+)(&.*)?$/) {
					unless ($shared) {
						$thread{'description'}.='<a href="'.entityencode(urldecode($3)).'">';
						$link=1;
					}
				}
				# there are two types of facebook user IDs
				elsif ($tag{'href'} =~ /^\/([A-Za-z0-9\.]+)\/?(\?.*)?$/) {
					my $person = $1;
					unless ($person =~ /^(photo|post).php$/) {
						if ($tag{'href'} =~ /^\/profile\.php\?(.*&)?id=([^&]+)(&.*)?$/) {
							$person = urldecode($2);
						}
						unless ($shared) {
							$thread{'description'}.='<b class="ni">'.(($$names{$person} ne '')?$$names{$person}:$$names{'default'});
							$link=1;
							$namelink=1;
							$hidename=1;
						}
						
						if ($thread{($shared?'shared-':'').'author'} eq '') {
							$thread{($shared?'shared-':'').'author'} = $person;
							$thread{($shared?'shared-':'').'name'} = ($$names{$person} ne '')?$$names{$person}:$$names{'default'};
						}
					}
				}
				elsif ($tag{'href'} eq '#') {
					unless ($shared) {
						$thread{'description'}.='<b class="ni">'.$$names{'default'};
						$link=1;
						$namelink=1;
						$hidename=1;
					}
					if ($thread{($shared?'shared-':'').'author'} eq '') {
						$thread{($shared?'shared-':'').'name'} = $$names{'default'};
					}
				}
			}
			elsif ($tag{'<'} eq '/a') {
				if($link) {
					if($namelink) {
						$thread{'description'}.='</b>';
						$namelink=0;
					}
					else {
						$thread{'description'}.='</a>';
					}
					$link=0;
					$hidename=0;
				}
			}
			# go out of <h3>
			elsif ($tag{'<'} eq '/h3') {
				$mode='thread';
			}
		}
		
		# the firstpost's content
		elsif ($mode eq 'thread-content') {
			# There should not be any sub<div>s. Ignore everything inside.
			if ($tag{'<'} eq 'div') {
				# if($tag{'class'} ne '') {
					# $level+=$level2+1;
					# $mode='thread';
				# }
				# else{
					++$level2;
					$ignoretext=1;
				# }
			}
			elsif ($tag{'<'} eq '/div') {
				if($level2){
					--$level2;
					unless($level2) {
						$ignoretext=0;
					}
				}
				else {
					$mode = 'thread';
					$content = 0;
				}
			}
			elsif ($tag{'<'} eq 'br') {
				$thread{($shared?'shared-':'').'postcontent'}.='<br>';
				$ignoretext=0;
			}
			elsif ($tag{'<'} eq 'p') {
				$thread{($shared?'shared-':'').'postcontent'}.='<p>';
				$ignoretext=0;
			}
			elsif ($tag{'<'} eq '/p') {
				$thread{($shared?'shared-':'').'postcontent'}.='</p>';
				$ignoretext=1;
			}
			elsif (!$ignoretext) {
				# inline image (smiley?)
				if (($tag{'<'} eq 'img')and($tag{'src'} =~ /^https?:\/\/([a-z0-9\.\-]+)?fbcdn\.net\/rsrc\.php\/(.*)$/)) {
					my $imgurl = urldecode($2);
					my $imgid = 'r_';
					$imgurl =~ s/([^A-Za-z0-9_\.])/sprintf ("@%02X",ord($1))/eg;
					
					while(length($imgurl)>240) {
						$imgid.=substr($imgurl,0,120).'-/';
						$imgurl=substr($imgurl,120);
					}
					$imgid.=$imgurl;
				
					my $imgkey = saveimg($tag{'src'},$imgid,$$settings{'id'});
					if ($imgkey ne '') {
						$thread{($shared?'shared-':'').'postcontent'}.='<img src="&img@i'.$imgid.'@k'.$imgkey.';" alt="'.$tag{'alt'}.'">';
					}
					else {
						$thread{($shared?'shared-':'').'postcontent'}.='<img src="" alt="'.$tag{'alt'}.'">';
					}
				}
				elsif ($tag{'<'} eq 'a') {
					# a link to an external page
					if ($tag{'href'} =~ /^https?:\/\/([a-z0-9\.\-]+)?facebook\.com\/l\.php\?(.*&)?u=([^&]+)(&.*)?$/) {
						$thread{($shared?'shared-':'').'postcontent'}.='<a href="'.entityencode(urldecode($3)).'">';
						$link=1;
					}
					# a link to a user
					elsif ($tag{'href'} =~ /^\/([A-Za-z0-9\.]+)\/?(\?.*)$/) {
						my $person=$1;
						if ($tag{'href'} =~ /^\/profile\.php\?(.*&)?id=([^&]+)(&.*)?$/) {
							$person = urldecode($2);
						}
						$thread{($shared?'shared-':'').'postcontent'}.='<b class="ni">'.(($$names{$person} ne '')?$$names{$person}:$$names{'default'});
						$link=1;
						$namelink=1;
						$hidename=1;
					}
					elsif ($tag{'href'} eq '#') {
						$thread{($shared?'shared-':'').'postcontent'}.='<b class="ni">'.$$names{'default'};
						$link=1;
						$namelink=1;
						$hidename=1;
					}
				}
				elsif ($tag{'<'} eq '/a') {
					if($link) {
						if($namelink) {
							$thread{($shared?'shared-':'').'postcontent'}.='</b>';
							$namelink=0;
						}
						else {
							$thread{($shared?'shared-':'').'postcontent'}.='</a>';
						}
						$link=0;
						$hidename=0;
					}
				}
				else {
				# $thread{'postcontent'}.='<!'.$tag{'<'}.'!>';
				}
			}
			# a link for "more..." outside the <p>s = past incomplete!
			elsif ($tag{'<'} eq 'a') {
				if (($tag{'href'}=~/^\/groups\/$$settings{'id'}\/?\?(.*&)?id=([^&]+)(&.*)?$/) and ($pagetype ne 'thread')) {
					unless($incomplete) {
						$thread{($shared?'shared-':'').'postcontent'}.='<p><b class = "br">Post not completely archived.</b></p>';
					}
					$incomplete=1;
				}
				
				### this code had to be duplicated:
				
				
				# there is an image attached
				elsif ($tag{'href'} =~ /^\/photo\.php\?(.*&)?fbid=([0-9]+)(&.*)?$/) {
					my $imgnum=$2;
					my $imgid='a_';
					
					while(length($imgnum)>240) {
						$imgid.=substr($imgnum,0,120).'-/';
						$imgnum=substr($imgnum,120);
					}
					$imgid.=$imgnum;
					
					++$attnumber;
					$thread{($shared?'shared-':'').'img-'.$attnumber}=$imgid;
					$mode = 'thread-attachment-img';
				}
				elsif ($tag{'href'} =~ /^\/[A-Za-z0-9\.]+\/photos\/([^\?]+)(\?.*)?$/) {
					my $imgurl = urldecode($1);
					my $imgid = 's_';
					$imgurl =~ s/([^A-Za-z0-9_\.])/sprintf ("@%02X",ord($1))/eg;
					
					while(length($imgurl)>240) {
						$imgid.=substr($imgurl,0,120).'-/';
						$imgurl=substr($imgurl,120);
					}
					$imgid.=$imgurl;
					
					++$attnumber;
					$thread{($shared?'shared-':'').'img-'.$attnumber}=$imgid;
					$mode = 'thread-attachment-img';
				}
				elsif ($tag{'href'} =~ /^\/video_redirect\/?\?(.*)$/) {
					my %linkcgi = getcgi($1);
					my $imgurl;
					my $imgid='v_';
					if ($linkcgi{'src'} =~ /^https?:\/\/([a-z0-9\.\-]+)?fbcdn\.net\/(hvideo[^\?]+)(\?.*)?$/) {
						$imgurl= $2;
						$imgurl =~ s/([^A-Za-z0-9_\.])/sprintf ("@%02X",ord($1))/eg;
						
						while(length($imgurl)>240) {
							$imgid.=substr($imgurl,0,120).'-/';
							$imgurl=substr($imgurl,120);
						}
						$imgid.=$imgurl;
						++$attnumber;
						$thread{($shared?'shared-':'').'img-'.$attnumber}=$imgid;
						$mode = 'thread-attachment-img';
					}
				}
				# there is a link attached
				elsif ($tag{'href'} =~ /^https?:\/\/([a-z0-9\.\-]+)?facebook\.com\/l\.php\?(.*&)?u=([^&]+)(&.*)?$/) {
					++$attnumber;
					$thread{($shared?'shared-':'').'link-'.$attnumber}=urldecode($3);
					$mode = 'thread-attachment-link';
				}
				# # if thread id is not known it can be determined from this link.
				# # also, the number of replies may be found in one of these links.
				# elsif ($tag{'href'} =~ /^\/groups\/$$settings{'id'}\/?\?(.*&)?id=([^&]+)(&.*)?$/) {
					# if ($thread{'id'} eq '') {
						# $thread{'id'} = $2;
						# print "Thread $thread{'id'}\n";
					# }
					# $mode = 'thread-replies';
				# }
				
				# end of duplicated code
			}
		}
		
		# time
		elsif ($mode eq 'thread-time') {
			if ($tag{'<'} eq '/abbr') {
				$mode = 'thread';
			}
		}
		
		# an attached image
		elsif ($mode eq 'thread-attachment-img') {
			if ($tag{'<'} eq 'img') {
				my $imgkey = saveimg($tag{'src'},$thread{($shared?'shared-':'').'img-'.$attnumber},$$settings{'id'});
				if ($imgkey ne '') {
					$thread{($shared?'shared-':'').'imgkey-'.$attnumber}=$imgkey;
				}
				else {
					delete $thread{($shared?'shared-':'').'img-'.$attnumber};
					--$attnumber;
				}
			}
			elsif ($tag{'<'} eq '/a') {
				$mode = ($content?'thread-content':'thread');
			}
		}
		
		#an attached link
		elsif ($mode eq 'thread-attachment-link') {
			#"title" is found in <h3> 
			if($tag{'<'} eq 'h3'){
				$mode = 'thread-attachment-link-title';
			}
			# an image included to illustrate the attached link. Facebook automatically chooses it...
			elsif (($tag{'<'} eq 'img')and($tag{'src'} =~ /^https?:\/\/([a-z0-9\.\-]+)?fbcdn\.net\/safe_image\.php\?(.*&)?url=([^&]+)(&.*)?$/)) {
				my $imgurl = urldecode($3);
				my $imgid='i_';
				$imgurl =~ s/([^A-Za-z0-9_\.])/sprintf ("@%02X",ord($1))/eg;
				
				while(length($imgurl)>240) {
					$imgid.=substr($imgurl,0,120).'-/';
					$imgurl=substr($imgurl,120);
				}
				$imgid.=$imgurl;
				my $imgkey = saveimg($tag{'src'},$imgid,$$settings{'id'});
				if ($imgkey ne '') {
					$thread{($shared?'shared-':'').'img-'.$attnumber}=$imgid;
					$thread{($shared?'shared-':'').'imgkey-'.$attnumber}=$imgkey;
				}
			}
			elsif ($tag{'<'} eq '/a') {
				$mode = ($content?'thread-content':'thread');
			}
		}
		
		elsif ($mode eq 'thread-attachment-link-title') {
			if ($tag{'<'} eq '/h3') {
				$mode = 'thread-attachment-link';
			}
		}
		
		elsif ($mode eq 'thread-replies') {
			if ($tag{'<'} eq '/a') {
				$mode = 'thread';
			}
		}
		
		elsif ($mode eq 'thread-likes') {
			if ($tag{'<'} eq '/a') {
				$mode = 'thread';
			}
		}
		
		# list of posts. look for <div>s with posts.
		elsif ($mode eq 'posts') {
			# post found
			if (($tag{'<'} eq 'div') and ($tag{'id'} =~ /^([0-9]+)$/)) {
				%post = ();
				
				$post{'id'} = $1;
				$post{'threadid'} = $threadid;
				$post{'groupid'} = $$settings{'id'};
				$post{'timenumber'} = $timenumber;
				
				$mode = 'post';
				$level=0;
				$attnumber=0;
				
				# if page type is 'post' the URL is not enough to see which group the
				# posts belongs to. has to be found on the page BEFORE entering first
				# post <div>!
				if($pagetype eq 'post') {
					if(!$groupid) {
						print "Can't determine if post belongs to group $$settings{'id'}.\n";
						$mode='';
						last;
					}
					elsif($post{'id'} eq $postid) {
						# In this page type it's important to check if this post is a reply
						# to a thread or to a post.
						$firstpost=1;
					}
					else {
						$firstpost=0;
						$post{'postid'} = $postid;
					}
				}
				print "Post ".((($pagetype eq 'post') and !$firstpost)?"$post{'postid'}/":"")."$post{'id'}\n";
			}
			# Link can be used to determine which group the post belongs to.
			elsif (($tag{'<'} eq 'a') and ($pagetype eq 'post') and ($tag{'href'} =~ /^\/groups\/([0-9]+)\/?\?/)) {
				if ($1 eq $$settings{'id'}) {
					$groupid = 1;
				}
				else {
					print "Post does not belong to group $$settings{'id'}.\n";
					$mode = '';
					last;
				}
			}
		}
		
		# one post
		elsif ($mode eq 'post') {
			# thread author is in first <h3>
			if ($tag {'<'} eq 'h3') {
				$mode = 'post-author';
			}
			elsif ($tag{'<'} eq 'abbr') {
				$mode = 'post-time';
			}
			if ($tag{'<'} eq 'div') {
				# Post content is (always?) in the first <div> without a class name after author
				if(($tag{'class'} eq '')and(!defined($post{'content'}))) {
					$mode = 'post-content';
					$level2=0;
					$ignoretext=0;
					$hidename=0;
					$link=0;
					$namelink=0;
				}
				else {
					++$level;
				}
			}
			elsif ($tag{'<'} eq '/div') {
				if ($level) {
					--$level;
				}
				else {
					# end of post <div>. Now prepare to save the file.
					# path depends if it's a post or a postreply.
					$mode = 'posts';
					
					my $postfile;
					my $postpath = ARCH_PATH.$$settings{'id'}.'/';
					unless (-d $postpath) {
						unless (mkdir $postpath) {
							print "Can't mkdir $postpath.\n";
						}
					}
					if(($pagetype eq 'post')and !$firstpost){
						$postpath.='postreply/';
					}
					else {
						$postpath.='post/';
					}
					unless (-d $postpath) {
						unless (mkdir $postpath) {
							print "Can't mkdir $postpath.\n";
						}
					}
					$postpath.=$post{'threadid'}.'/';
					unless (-d $postpath) {
						unless (mkdir $postpath) {
							print "Can't mkdir $postpath.\n";
						}
					}
					if(($pagetype eq 'post')and !$firstpost){
						$postpath.=$post{'postid'}.'/';
						unless (-d $postpath) {
							unless (mkdir $postpath) {
								print "Can't mkdir $postpath.\n";
							}
						}
					}
					
					$postpath.=$post{'id'};
					
					if (sysopen ($postfile, $postpath, O_RDWR | O_CREAT)) {
						if (flock ($postfile, 2)) {
							%post2 = readdatafile($postfile);
							
							# Don't overwrite newer information with older.
							if (($post2{'timenumber'} ne '')and($post2{'timenumber'}>=$post{'timenumber'})) {
								print ("Newer version already saved.\n\n");
							}
							else {
								# delete previous information about attachments - the numbers
								# could have changed.
								foreach my $ind (keys %post2) {
									if($ind =~ /^((img(key)?)|(link(text|title)?))-[0-9]+$/) {
										delete $post2{$ind};
									}
								}
								# overwrite previous information with new one
								foreach my $ind (keys %post) {
									$post2{$ind}=$post{$ind};
								}
								if ($post2{'key'} eq '') {
									$post2{'key'} = key(KEY_BITS);
								}
								# write data to file
								if (seek($postfile, 0, 0)) {
									writedatafile($postfile,%post2);
									truncate ($postfile , tell($postfile));
							
									if(DETAILED_LOGS) {
										foreach my $ind (keys %post2) {
											print "$ind: $post2{$ind}\n";
										}
									}
									print "saved.\n\n";
								}
								else {
									print "Can't seek $postfile.\n\n";
								}
							}
						}
						else {
							print "Can't lock $postfile.\n\n";
						}
						close ($postfile);
					}
					else
					{
						print "Can't open $postpath.\n\n";
					}
				}
			}
			
			elsif ($tag{'<'} eq 'a') {
				if (lc($tag{'aria-label'})eq 'likes') {
					$mode='post-likes';
				}
				# there is an image attached
				elsif ($tag{'href'} =~ /^\/photo\.php\?(.*&)?fbid=([0-9]+)(&.*)?$/) {
					my $imgnum=$2;
					my $imgid='a_';
					
					while(length($imgnum)>240) {
						$imgid.=substr($imgnum,0,120).'-/';
						$imgnum=substr($imgnum,120);
					}
					$imgid.=$imgnum;
					
					++$attnumber;
					$post{'img-'.$attnumber}=$imgid;
					$mode = 'post-img';
				}
				# there is a link attached
				elsif ($tag{'href'} =~ /^https?:\/\/([a-z0-9\.\-]+)?facebook\.com\/l\.php\?(.*&)?u=([^&]+)(&.*)?$/) {
					++$attnumber;
					$post{'link-'.$attnumber}=urldecode($3);
					$mode = 'post-link';
				}
				# the number of replies may be found in one of these links.
				elsif ($tag{'href'} =~ /^\/comment\/replies\/?\?/) {
					$mode = 'post-replies';
				}
			}
		}
		
		# an attached image
		elsif ($mode eq 'post-img') {
			if ($tag{'<'} eq 'img') {
				my $imgkey = saveimg($tag{'src'},$post{'img-'.$attnumber},$$settings{'id'});
				if ($imgkey ne '') {
					$post{'imgkey-'.$attnumber}=$imgkey;
				}
				else {
					delete $post{'img-'.$attnumber};
					--$attnumber;
				}
			}
			elsif ($tag{'<'} eq '/a') {
				$mode = 'post';
			}
		}
		
		#an attached link
		elsif ($mode eq 'post-link') {
			#"title" is found in <h3> 
			if($tag{'<'} eq 'h3'){
				$mode = 'post-link-title';
			}
			# an image included to illustrate the attached link. Facebook automatically chooses it...
			elsif (($tag{'<'} eq 'img')and($tag{'src'} =~ /^https?:\/\/([a-z0-9\.\-]+)?fbcdn\.net\/safe_image\.php\?(.*&)?url=([^&]+)(&.*)?$/)) {
				my $imgurl = urldecode($3);
				my $imgid='i_';
				$imgurl =~ s/([^A-Za-z0-9_\.])/sprintf ("@%02X",ord($1))/eg;
				
				while(length($imgurl)>240) {
					$imgid.=substr($imgurl,0,120).'-/';
					$imgurl=substr($imgurl,120);
				}
				$imgid.=$imgurl;
				my $imgkey = saveimg($tag{'src'},$imgid,$$settings{'id'});
				if ($imgkey ne '') {
					$post{'img-'.$attnumber}=$imgid;
					$post{'imgkey-'.$attnumber}=$imgkey;
				}
			}
			elsif ($tag{'<'} eq '/a') {
				$mode = 'post';
			}
		}
		
		elsif ($mode eq 'post-link-title') {
			if ($tag{'<'} eq '/h3') {
				$mode = 'post-link';
			}
		}
		
		elsif ($mode eq 'post-author') {
			# name can be found in hyperlinks
			if ($tag{'<'} eq 'a') {
				# there are two types of facebook user IDs
				if ($tag{'href'} =~ /^\/([A-Za-z0-9\.]+)\/?(\?.*)?$/) {
					my $author = $1;
					if ($tag{'href'} =~ /^\/profile\.php\?(.*&)?id=([^&]+)(&.*)?$/) {
						$author = urldecode($2);
					}
					if ($post{'author'} eq '') {
						$post{'author'} = $author;
						$post{'name'} = ($$names{$author} ne '')?$$names{$author}:$$names{'default'};
					}
				}
			}
			elsif ($tag{'<'} eq '/h3') {
				$mode='post';
			}
		}
		
		# the post content
		elsif ($mode eq 'post-content') {
			# There should not be any sub<div>s. Ignore everything inside.
			if ($tag{'<'} eq 'div') {
				++$level2;
				$ignoretext=1;
			}
			elsif ($tag{'<'} eq '/div') {
				if($level2){
					--$level2;
					unless($level2){
						$ignoretext=0;
					}
				}
				else {
					$mode = 'post';
				}
			}
			elsif ($tag{'<'} eq 'br') {
				$post{'content'}.='<br>';
			}
			elsif ($tag{'<'} eq 'p') {
				$post{'content'}.='<p>';
			}
			elsif ($tag{'<'} eq '/p') {
				$post{'content'}.='</p>';
			}
			elsif (!$ignoretext) {
				if (($tag{'<'} eq 'img')and($tag{'src'} =~ /^https?:\/\/([a-z0-9\.\-]+)?fbcdn\.net\/rsrc\.php\/(.*)$/)) {
					my $imgurl = urldecode($2);
					my $imgid = 'r_';
					$imgurl =~ s/([^A-Za-z0-9_\.])/sprintf ("@%02X",ord($1))/eg;
					
					while(length($imgurl)>240) {
						$imgid.=substr($imgurl,0,120).'-/';
						$imgurl=substr($imgurl,120);
					}
					$imgid.=$imgurl;
				
					my $imgkey = saveimg($tag{'src'},$imgid,$$settings{'id'});
					if ($imgkey ne '') {
						$post{'content'}.='<img src="&img@i'.$imgid.'@k'.$imgkey.';" alt="'.$tag{'alt'}.'">';
					}
					else {
						$post{'content'}.='<img src="" alt="'.$tag{'alt'}.'">';
					}
				}
				elsif ($tag{'<'} eq 'a') {
					# a link to an external page
					if ($tag{'href'} =~ /^https?:\/\/([a-z0-9\.\-]+)?facebook\.com\/l\.php\?(.*&)?u=([^&]+)(&.*)?$/) {
						$post{'content'}.='<a href="'.entityencode(urldecode($3)).'">';
						$link=1;
					}
					# a link to a user
					elsif ($tag{'href'} =~ /^\/([A-Za-z0-9\.]+)\/?(\?.*)?$/) {
						my $person = $1;
						if ($tag{'href'} =~ /^\/profile\.php\?(.*&)?id=([^&]+)(&.*)?$/) {
							$person = urldecode($2);
						}
						$post{'content'}.='<b class="ni">'.(($$names{$person} ne '')?$$names{$person}:$$names{'default'});
						$link=1;
						$namelink=1;
						$hidename=1;
					}
					elsif ($tag{'href'} eq '#') {
						$post{'content'}.='<b class="ni">'.$$names{'default'};
						$link=1;
						$namelink=1;
						$hidename=1;
					}
				}
				elsif ($tag{'<'} eq '/a') {
					if($link) {
						if ($namelink) {
							$post{'content'}.='</b>';
							$namelink=0;
						}
						else {
							$post{'content'}.='</a>';
						}
						$link=0;
						$hidename=0;
					}
				}
				else {
				# $post{'content'}.='<!'.$tag{'<'}.'!>';
				}
			}
			
		}
		
		elsif ($mode eq 'post-time') {
			if ($tag{'<'} eq '/abbr') {
				$mode = 'post';
			}
		}
		
		elsif ($mode eq 'post-replies') {
			if ($tag{'<'} eq '/a') {
				$mode = 'post';
			}
		}
		
		elsif ($mode eq 'post-likes') {
			if ($tag{'<'} eq '/a') {
				$mode = 'post';
			}
		}
		
		# dealing with the tag is finished.
		
		# if the tag ends with "/>" chande it into an "</" tag ang go through it
		# again from the start of the loop
		if ($tag{"\\"} ne '') {
			$closetag = 1;
			next;
		}
		
		# now get the text between tags
		
		local $/ = '<';
		unless (defined ($text = <$contentfile>)) {
			close($contentfile);
			return;
		}
		local $/ = "\n";
		$text =~ s/<$//;
		
		# # DEBUG
		# if ($pagetype eq 'thread') {
			# print ">>$mode: $text\n";
		# }
		
		# depending on state add text to relevant fields.
		
		if($mode eq 'thread-content') {
			unless ($ignoretext or $hidename){
				$thread{($shared?'shared-':'').'postcontent'}.=$text;
			}
		}
		if($mode eq 'thread-author') {
			unless ($hidename or $shared){
				$thread{'description'}.=$text;
				if (($text =~ /[a-z]/) and (!$link) and ($description == 0)) {
					$description = 1;
				}
			}
		}
		# the format facebook uses for showing time is not always helpful (for
		# example: "2 mins ago") There is no setting to change it in facebook.
		# at least in the m.facebook.com. The bot corrently DOES NOT interpret the
		# text.
		elsif ($mode eq 'thread-time') {
			$thread{($shared?'shared-':'').'timetext'}.=facebooktime($timenumber,$text);
		}
		elsif ($mode eq 'thread-attachment-link-title') {
			$thread{($shared?'shared-':'').'linktitle-'.$attnumber}.=$text;
		}
		elsif ($mode eq 'thread-attachment-link') {
			$thread{($shared?'shared-':'').'linktext-'.$attnumber}.=$text;
		}
		elsif ($mode eq 'thread-replies') {
			if(lc($text) =~ /^[ \t\r\n]*([0-9]+)[ \t\r\n]+comments?/) {
				$thread{'replies'} = $1;
			}
		}
		elsif ($mode eq 'thread-likes') {
			if(lc($text) =~ /^[ \t\r\n]*([0-9]+)[ \t\r\n]*?/) {
				$thread{'likes'} = $1;
			}
		}
		
		if($mode eq 'post-content') {
			unless ($ignoretext or $hidename){
				$post{'content'}.=$text;
			}
		}
		elsif ($mode eq 'post-time') {
			$post{'timetext'}.=facebooktime($timenumber,$text);
		}
		elsif ($mode eq 'post-link-title') {
			$post{'linktitle-'.$attnumber}.=$text;
		}
		elsif ($mode eq 'post-link') {
			$post{'linktext-'.$attnumber}.=$text;
		}
		elsif ($mode eq 'post-replies') {
			if(lc($text) =~ /^[ \t\r\n]*([0-9]+)[ \t\r\n]+repl(y|ies)/) {
				$post{'replies'} = $1;
			}
		}
		elsif ($mode eq 'post-likes') {
			if(lc($text) =~ /^[ \t\r\n]*([0-9]+)[ \t\r\n]*?/) {
				$post{'likes'} = $1;
			}
		}
	}
	close ($contentfile);
}

# Function to get information about a html tag.
# The argument is the tag text without the <>!
# It returns a hash with the attributes' values.
# special values:
# '<' - the tag name (may start with '/'),
# '/' - when the tag starts with '/',
# '\' - when the tag ends with '/',
sub taginfo {
	(my $tagtext) = @_;
	my %tag;
	
	# if ($tagtext =~ /^div/) {
		# print "$tagtext\n";
	# }
	
	if ($tagtext =~ /^((\/?)[^ \t\r\n<>\/"=]+)([ \t\r\n].*)?$/) {
		$tag{'<'}=$1;
		if ($2 ne '') {
			$tag{'/'} = '/';
		}
		$tagtext = $3;
	}
	else {
		return %tag;
	}
	
	while ($tagtext =~ /^[ \t\r\n]*([^ \t\r\n<>\/"=]+)[ \t\r\n]*=[ \t\r\n]*"([^"]+)"([ \t\r\n].*)?$/) {
		$tag{$1} = entitydecode($2);
		$tagtext = $3;
	}
	
	if ($tagtext =~ /\/[ \t\r\n]*$/) {
		$tag{"\\"}="\\";
	}
	return %tag; 
}

# The function to save an image from proxy archive
# Arguments:
# 1 - url of image
# 2 - image id
# 3 - group id
# Returns 1 on success and 0 on failure.
#
sub saveimg {
	(my $url, my $id, my $groupid) = @_;
	
	my $basepath;
	my $imgpath;
	my $headpath;
	my $headopen=0;;
	my $headfile;
	my %header;
	
	my $archimgpath;
	my $archheadpath;
	my $archbasepath;
	my %archheader;
	
	(my $prot, my $host, my $port, my $path, my $query);
	my @stat;
	
	print "Image $id\n";
	
	# prepare path
	$basepath=ARCH_PATH.$groupid.'/';
	unless (-d $basepath){
		unless (mkdir $basepath) {
			print "Can't mkdir $basepath.\n";
			return '';
		}
	}
	$basepath.='image/';
	unless (-d $basepath){
		unless (mkdir $basepath) {
			print "Can't mkdir $basepath.\n";
			return '';
		}
	}
	
	my $idtemp = $id;
	while((my $ind = index($idtemp,'/'))>=0) {
		$basepath.=substr($idtemp, 0, $ind+1);
		$idtemp=substr($idtemp,$ind+1);
		unless (-d $basepath){
			unless (mkdir $basepath) {
				print "Can't mkdir $basepath.\n";
				return '';
			}
		}
	}
	
	$basepath.=$idtemp;
	$headpath=$basepath.'@h';
	$imgpath=$basepath.'@v';
	
	# check if image already saved
	if ( -f $imgpath) {
		if (open($headfile,"+<",$headpath)) {
			if (flock ($headfile, 2)) {
				%header = readheaderfile($headfile);
				if ($header{'key'} ne '') {
					print "Already saved.\n\n";
					close($headfile);
					return $header{'key'};
				}
				else {
					$headopen = 1;
				}
			}
			else {
				print "Can't lock $headpath.\n\n";
				close($headfile);
				return '';
			}
		}
	}
	
	($prot, $host, $port, $path, $query) = divideurl($url);
	$archbasepath = urldiv2path($prot, $host, $port, $path, $query);
	$archheadpath = $archbasepath.'@h';
	$archimgpath = $archbasepath.'@v';
	if (DETAILED_LOGS) {
		print "url: $url\n";
	}
	# Read the http header. Only interested in the status. If a redirection then
	# follow it and read again. If 200 it's ok to continue processing. Otherwise
	# return.
	for (my $ind=0; $ind<MAX_REDIRECTIONS; ++$ind) {
		%archheader = readheaderfile($archheadpath); 
		
		if ($archheader{'status'} =~ /^30[1237] /) {
			my $location;
			unless (defined($location = $archheader{'location'})) {
				last;
			}
			if ($location !~ /^[a-z]+:\/\//) {
				$location = $prot.'://'.$host.(($port ne '')?(':'.$port):'').$location;
			}
			if (DETAILED_LOGS) {
				print "Redirect: $location\n";
			}
			($prot, $host, $port, $path, $query) = divideurl($location);
			$archbasepath = urldiv2path($prot, $host, $port, $path, $query);
			$archheadpath = $archbasepath.'@h';
			$archimgpath = $archbasepath.'@v';
		}
		else {
			last;
		}
	}
	if ($archheader{'status'} !~ /^200 /) {
		print "Not found ($archheader{'status'}).\n\n";
		if ($headopen) {
			close ($headfile);
		}
		return '';
	}
	
	# save content-type and content-disposition information.
	foreach my $ind (keys %archheader) {
		if ($ind =~ /^(content-type|content-disposition)$/) {
			$header{$ind}=$archheader{$ind};
		}
	}
	# generate key
	$header{'key'}=key(KEY_BITS);
	# save image size
	if (@stat = stat($archimgpath)){
		$header{'content-length'}=$stat[7];
	}
	else {
		print "Can't stat $imgpath. $stat[7]\n";
	}
	$header{'id'}=$id;
	$header{'groupid'}=$groupid;
	$header{'timenumber'}=strftime('%Y%m%d%H%M%S',gmtime(time()));
	$header{'last-modified'}=strftime('%a, %d %b %Y %H:%M:%S GMT',gmtime(time()));
	
	if ($headopen) {
		unless (seek($headfile,0,0)) {
			print "Can't seek $headpath.\n\n";
			close ($headfile);
			return '';
		}
	}
	else{
		unless (open($headfile,">",$headpath)) {
			print "Can't open $headpath.\n\n";
			return '';
		}
		unless (flock ($headfile, 2)) {
			print "Can't lock $headpath.\n\n";
			close($headfile);
			return '';
		}
	}
	
	# save header
	writedatafile($headfile,%header);
	print $headfile "\n";
	close ($headfile);
	
	# Finally, copy the image.
	unless (copy($archimgpath,$imgpath)) {
		print "Can't copy $archimgpath.\n\n";
		return '';
	}
	print "saved.\n\n";
	return $header{'key'};
	
}

# function to get time from facebook time text
# arguments: 1 - time number (when page was downloaded), 2 - facebook time text
# returns a time number on success and the unchanged time text on failure.
sub facebooktime {
	(my $timenumber, my $timetext) = @_;
	$timetext = lc ($timetext);
	
	my $year;
	my $month;
	my $day;
	my $hour;
	my $minute;
	my $ampm;
	
	my $subtype;
	my $subvalue;
	my $returnvalue;
	
	# There are multible possible format of facebook time text format
	
	# Jan 2
	# January 2
	if ($timetext =~ /^ *([a-z]{3,}) +([0-9]+) *$/) {
		$month = $1;
		$day = int($2);
		$year = int(substr($timenumber,0,4));
		$hour='';
		$minute='';
	}
	# Jan 2, 2015
	# January 2, 2015
	elsif ($timetext =~ /^ *([a-z]{3,}) +([0-9]+), +([0-9]{4}) *$/) {
		$month = $1;
		$day = int($2);
		$year = int($3);
		$hour='';
		$minute='';
	}
	# Jan 2 at 10:02am
	# January 2 at 10:02am
	elsif ($timetext =~ /^ *([a-z]{3,}) +([0-9]+) +at +([0-9]{1,2}):([0-9]{2})([ap]m) *$/) {
		$month = $1;
		$day = int($2);
		$year = int(substr($timenumber,0,4));
		$hour=int($3);
		$minute=int($4);
		$ampm=$5;
		if($ampm eq 'pm') {
			$hour += 12;
		}
	}
	# Jan 2, 2015 at 10:02am
	# January 2, 2015 at 10:02am 
	elsif ($timetext =~ /^ *([a-z]{3,}) +([0-9]+), +([0-9]{4}) +at +([0-9]{1,2}):([0-9]{2})([ap]m) *$/) {
		$month = $1;
		$day = int($2);
		$year = int($3);
		$hour=int($4);
		$minute=int($5);
		$ampm=$6;
		if($ampm eq 'pm') {
			$hour += 12;
		}
	}
	# 1 hr
	# 2 hours ago
	# etc...
	elsif ($timetext =~ /^ *([0-9]+) +h(ou)?rs?( +ago)? *$/) {
		$subtype='h';
		$subvalue=int($1);
		$hour = int(substr($timenumber,8,2));
	}
	# 4 mins
	# 2 minutes ago
	# etc...
	elsif ($timetext =~ /^ *([0-9]+) +min(ute)?s?( +ago)? *$/) {
		$subtype='m';
		$subvalue=int($1);
		$minute = int(substr($timenumber,10,2));
		$hour = int(substr($timenumber,8,2));
	}
	# 2 days ago
	# etc...
	elsif ($timetext =~ /^ *([0-9]+) +days?( +ago)? *$/) {
		$subtype='d';
		$subvalue=int($1);
	}
	elsif ($timetext =~ /^ *yesterday +at +([0-9]{1,2}):([0-9]{2})([ap]m) */) {
		$subtype='d';
		$subvalue=1;
		$hour=int($1);
		$minute=int($2);
		$ampm=$3;
		if($ampm eq 'pm') {
			$hour += 12;
		}
	}
	elsif ($timetext =~ /^ *today +at +([0-9]{1,2}):([0-9]{2})([ap]m) */) {
		$hour=int($1);
		$minute=int($2);
		$ampm=$3;
		if($ampm eq 'pm') {
			$hour += 12;
		}
	}
	else {
		# print "FAIL 1 -".urlencode($timetext)."-\n";
		return $timetext;
	}
	
	if ($year !~ /^[0-9]+$/) {
		$year = int(substr($timenumber,0,4));
	}
	if ($month !~ /^[0-9]+$/) {
		if ($month eq '') {
			$month = int(substr($timenumber,4,2));
		}
		elsif ($month =~ /^jan/) {
			$month=1;
		}
		elsif ($month =~ /^feb/) {
			$month=2;
		}
		elsif ($month =~ /^mar/) {
			$month=3;
		}
		elsif ($month =~ /^apr/) {
			$month=4;
		}
		elsif ($month =~ /^may/) {
			$month=5;
		}
		elsif ($month =~ /^jun/) {
			$month=6;
		}
		elsif ($month =~ /^jul/) {
			$month=7;
		}
		elsif ($month =~ /^aug/) {
			$month=8;
		}
		elsif ($month =~ /^sep/) {
			$month=9;
		}
		elsif ($month =~ /^oct/) {
			$month=10;
		}
		elsif ($month =~ /^nov/) {
			$month=11;
		}
		elsif ($month =~ /^dec/) {
			$month=12;
		}
		else {
			# print "FAIL 2\n";
			return $timetext;
		}
		
	}
	if ($day !~ /^[0-9]+$/) {
		$day = substr($timenumber,6,2);
	}
	
	if ($subtype eq 'm') {
		$minute -= $subvalue;
		$subvalue = 0;
		while($minute <0) {
			$minute += 60;
			++$subvalue;
			$subtype = 'h';
		}
	}
	if ($subtype eq 'h') {
		$hour -= $subvalue;
		$subvalue = 0;
		while($hour <0) {
			$hour += 24;
			++$subvalue;
			$subtype = 'd';
		}
	}
	if ($subtype eq 'd') {
		$day -= $subvalue;
		if ($day <1) {
			--$month;
			if($month<1) {
				--$year;
				$month=12;
				$day = 31;
			}
			else {
				if ($month == 2) {
					if($year%4) {
						$day = 28;
					}
					elsif ($year%100) {
						$day = 29;
					}
					elsif ($year%400) {
						$day = 28;
					}
					else {
						$day = 29;
					}
				}
				elsif ((($month%2) and ($month < 8)) or ((!($month%2)) and ($month >= 8))) {
					$day = 31;
				}
				else {
					$day = 30;
				}
			}
		}
	}
	
	$returnvalue= sprintf('%04u%02u%02u',$year,$month,$day);
	if($hour ne '') {
		$returnvalue.=sprintf('%02u',$hour);
		if($minute ne '') {
			$returnvalue.=sprintf('%02u',$minute);
		}
	}
	return $returnvalue;
}