#!/usr/bin/perl

# configure.pl
# 13.01.2016
#
# The facebook interface software, when run on a server, will use different
# directories, host names, tcp ports, etc. than the server on which this
# software was originally written.
# These things are defined in the file 'settings'.
# This script is called from the makefile. It reads the settings file and
# inserts the information in the source files.
#
#    Copyright (C) 2015-2016  Balthasar Szczepański
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

unless ($ARGV[0]) {
	print STDERR "Configfile missing.\n";
	exit 1;
}

unless (open $configfile, "<", $ARGV[0]) {
	print STDERR "Cannot open configfile\n";
	exit 2;
}

# Read the config file, line format:
# some_name = some value # some comment
while (defined(my $line = <$configfile>)) {
	$line =~ s/[\r\n]//g;
	$line =~ s/#.*$//; #comment
	if ($line =~ /^[ \t]*([a-zA-Z0-9_\-\.]+)[ \t]*=[ \t]*([^ \t](.*[^ \t])?)[ \t]*$/){
		my $name=$1;
		my $value=$2;
		$set{$name}=$value;
	}
}
close ($configfile);

# Now generate things to be inserted.

$def{'PROXY_ARCH_PATH'}     = "use constant PROXY_ARCH_PATH    => '".$set{'proxy_data_path'}."archive/';";
$def{'ARCH_PATH'}           = "use constant ARCH_PATH          => '".$set{'data_path'}."group/';";
$def{'GROUPSETTINGS_PATH'}  = "use constant GROUPSETTINGS_PATH => '".$set{'data_path'}."groupsettings/';";
$def{'KEY_BITS'}            = "use constant KEY_BITS           => ".$set{'key_bits'}.";";
$def{'THREADS_IN_PAGE'}     = "use constant THREADS_IN_PAGE    => ".$set{'threads_in_page'}.";";
$def{'POSTS_IN_PAGE'}       = "use constant POSTS_IN_PAGE      => ".$set{'posts_in_page'}.";";
$def{'MAX_REDIRECTIONS'}    = "use constant MAX_REDIRECTIONS   => ".$set{'max_redirections'}.";";
$def{'IF_CSS_PATH'}         = "use constant IF_CSS_PATH        => '".$set{'interface_path'}."/if.css';";
$def{'INTERFACE_PATH'}      = "use constant INTERFACE_PATH     => '".$set{'interface_path'}."/view';";
$def{'LOGIN_PATH'}          = "use constant LOGIN_PATH         => '".$set{'interface_path'}."/view/login';";
$def{'LOGOUT_PATH'}         = "use constant LOGOUT_PATH        => '".$set{'interface_path'}."/view/logout';";
$def{'LIST_PATH'}           = "use constant LIST_PATH          => '".$set{'interface_path'}."/view/list';";
$def{'GROUP_PATH'}          = "use constant GROUP_PATH         => '".$set{'interface_path'}."/view/group';";
$def{'THREAD_PATH'}         = "use constant THREAD_PATH        => '".$set{'interface_path'}."/view/thread';";
$def{'POST_PATH'}           = "use constant POST_PATH          => '".$set{'interface_path'}."/view/post';";
$def{'IMAGE_PATH'}          = "use constant IMAGE_PATH         => '".$set{'interface_path'}."/view/image';";
$def{'PASS_PATH'}           = "use constant PASS_PATH          => '".$set{'data_path'}."pass/';";
$def{'ACCESS_PATH'}         = "use constant ACCESS_PATH        => '".$set{'data_path'}."access/';";
$def{'TIMEOUT_UNLOCK'}      = "use constant TIMEOUT_UNLOCK     => ".$set{'timeout_unlock'}.";";
$def{'TIMEOUT_INACT'}       = "use constant TIMEOUT_INACT      => ".$set{'timeout_inact'}.";";
$def{'HTML_HEAD'}           = "use constant HTML_HEAD          => '".$set{'html_head'}."';";
$def{'HTML_TOP'}            = "use constant HTML_TOP           => '".$set{'html_top'}."';";
$def{'HTML_BOTTOM'}         = "use constant HTML_BOTTOM        => '".$set{'html_bottom'}."';";

if(($set{'detailed_logs'} eq 'yes') or ($set{'detailed_logs'} =~ /^[0-9]*[1-9][0-9]*$/)) {
	$def{'DETAILED_LOGS'}     = "use constant DETAILED_LOGS      => 1;";
}
else {
	$def{'DETAILED_LOGS'}     = "use constant DETAILED_LOGS      => 0;";
}


$def{'PROXY_LIB'}   = "use lib '".$set{'proxy_lib_path'}."';";
$def{'FACEBUG_LIB'} = "use lib '".$set{'lib_path'}."';";

$def{'PATH'} = "\$ENV{'PATH'} = '".$set{'path'}."';";


$def{'PERL'} = "#!".$set{'perl'};

$def{'INTERFACE_PL'}        = '#define INTERFACE_PL       "'.$set{'bin_path'}.'interface.pl"';
$def{'INTERFACE_PL_ERRLOG'} = '#define INTERFACE_PL_ERRLOG "'.$set{'log_path'}.'interface-stderr.log"';

$wwwpath = $set{'www_path'};
$wwwpath =~ s/\/$//;

$def{'CGI_ALIAS'}  = 'ScriptAlias '.$set{'interface_path'}.'/view '.$set{'bin_path'}.'interface';
$def{'PATH_ALIAS'} = 'Alias       '.$set{'interface_path'}.'      '.$wwwpath; 

$def{'BOT_CRONTAB'}       = $set{'bot_crontab'}.' '.$set{'bin_path'}.'bot'.(($set{'bot_args'} ne '')?(' '.$set{'bot_args'}):'').' >> '.$set{'log_path'}.'bot.log';
$def{'RM_ACCESS_CRONTAB'} = $set{'rm_access_crontab'}.' '.$set{'bin_path'}.'rmaccess';
$def{'OLDLOGS_CRONTAB'}   = $set{'oldlogs_crontab'}.' '.$set{'proxy_bin_path'}.'oldlogs '.$set{'log_path'}.' '.$set{'log_size_limit'}.' '.$set{'logs_total'}.' '.$set{'logs_uncompressed'};

$def{'CC'} = 'CC='.$set{'gcc'};
$def{'CF'} = 'CF='.$set{'c_flags'};
$def{'PL'} = 'PL='.$set{'perl'};
$def{'MV'} = 'MV='.$set{'mv'};
$def{'CP'} = 'CP='.$set{'cp'};
$def{'RM'} = 'RM='.$set{'rm'};
$def{'OD'} = 'OD='.$set{'bin_path'};
$def{'WD'} = 'WD='.$set{'www_path'};
$def{'LD'} = 'LD='.$set{'lib_path'};
$def{'CM'} = 'CM='.$set{'chmod'};


# Now go through input file, find lines to be replaced. Format:
# ###SOME_NAME;
# If found - replace.

while (defined($line = <STDIN>)) {
	$line =~ s/[\r\n]//g;
	if ($line =~ /###([a-zA-Z0-9_]+);/) {
		print "$def{$1}\n";
	}
	else {
		print "$line\n";
	}
}
